#!/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision: 5366 $
$LastChangedDate:: 2021-12-20 17:55:34 +0100#$

This file is part of SyMSpace.

Copyright (c) 2020, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""
import sys
import os
import glob
import json
import threading
import subprocess
import tempfile
import pathlib
import ast
import psutil

from qtpy import QtGui
from qtpy.QtCore import Qt, QUrl
from qtpy.QtWidgets import (QTabWidget, QCheckBox, QApplication, QGridLayout, QGroupBox, QTreeWidgetItemIterator,
                            QWidget, QPushButton, QTreeWidgetItem, QTreeWidget, QErrorMessage,
                            QVBoxLayout, QMenu, QLineEdit, QLabel, QComboBox)
from qtpy.QtWebEngineWidgets import QWebEngineView


def showErrorInGUI(ex: Exception):
    app = QApplication([])
    error_dialog = QErrorMessage()
    error_dialog.showMessage(str(ex))
    sys.exit(app.exec_())


try:
    symspace_root = os.getenv('MagOpt_Home')
    cs_root = os.getenv('ComponentSpace')
    if symspace_root is None:
        raise EnvironmentError('Environment Variable "MagOpt_Home" is not defined!')
    elif not os.path.isdir(symspace_root):
        raise NotADirectoryError('"MagOpt_Home" is not a directory!')
    if cs_root is None:
        raise EnvironmentError('Environment Variable "ComponentSpace" is not defined!')
    elif not os.path.isdir(cs_root):
        raise NotADirectoryError('"ComponentSpace" is not a directory!')

    system_python = os.path.join(symspace_root, 'System', 'Python', 'src')
    general_python = os.path.join(symspace_root, 'System', 'General', 'Python', 'src')

    cs_src = os.path.join(cs_root, 'src')
    if system_python not in sys.path:
        sys.path.append(system_python)
    if general_python not in sys.path:
        sys.path.append(general_python)
    if cs_src not in sys.path:
        sys.path.append(cs_src)

    import symspace.testing.junit2html as junit2html
    from symspace.gui.qt import set_QT_PLUGIN_PATH
    from symspace.osutils.processutils import csymspaceRun
    from symspace.osutils.processutils import pythonRun
    from symspace.utils.decorators import try_catch

    from componentspace.create_buildscript import ComponentBuilder
    from componentspace.testing.create_tests import ComponentTestBuilder
    from componentspace.testing.project_tests import ProjectTestBuilder
    from componentspace.help.developer_help import DeveloperHelp
    import componentspace.osutils.fileutils as fu
    from componentspace.gui import external_functions

    set_QT_PLUGIN_PATH()

except Exception as pre_ex:
    showErrorInGUI(pre_ex)


class ComponentSpaceCenterQt(QWidget):
    ncols = 5
    TreeCol, VersionCol, UserGroupCol, DependenciesCol, PackagesCol = range(ncols)

    @staticmethod
    def __getTempDir() -> str:
        tmp_dir = os.path.join(tempfile.gettempdir(), 'tmp_ComponentSpaceGUI')
        return tmp_dir

    @classmethod
    def __makeTempDir(cls) -> str:
        tmp_dir = cls.__getTempDir()
        if not os.path.isdir(tmp_dir):
            os.makedirs(tmp_dir)
        return tmp_dir

    @classmethod
    def __writeJSONFile(cls, obj):
        tmp = tempfile.NamedTemporaryFile(dir=cls.__makeTempDir(), mode='w+t', suffix='.json', delete=False)
        json_file = tmp.name
        with open(json_file, 'w') as fh:
            json.dump(obj, fh, indent=4, sort_keys=False)
        return tmp.name

    def __init__(self):
        # noinspection PyArgumentList
        super().__init__()
        self.script = external_functions.__file__
        self.info_file = os.path.join(cs_root, 'component_info.json')
        self.icon_path = os.path.join(cs_root, 'src', 'componentspace', 'gui', 'Icon', 'Icon_128x128.png')
        self.title = 'ComponentSpace Control Center'
        # self.left = 10
        # self.top = 10
        self.width = 1600
        self.height = 900
        self.tree_directories = {}
        self.tree_components = {}
        self.excel_tree_directories = {}
        self.excel_tree_files = {}

        self.tabs = QTabWidget()
        # noinspection PyArgumentList
        self.tab1 = QWidget()
        # noinspection PyArgumentList
        self.tab2 = QWidget()
        # noinspection PyArgumentList
        self.tab3 = QWidget()
        # self.tabs.resize(300, 200)

        self.tree = QTreeWidget()
        self.tree_control_box = QGroupBox('Tree Control')
        self.control_box = QGroupBox('Component Control')
        self.browser1_box = QGroupBox('Browser')

        self.excel_tree = QTreeWidget()
        self.excel_tree_control_box = QGroupBox('Tree Control')
        self.excel_box = QGroupBox('Excel Tests')
        self.browser2_box = QGroupBox('Browser')

        self.help_box = QGroupBox('Help')
        self.browser3_box = QGroupBox('Browser')

        self.tab1_browser = QWebEngineView()
        self.tab2_browser = QWebEngineView()
        self.tab3_browser = QWebEngineView()

        self.field_component_timeout = QLineEdit('300.0')
        self.field_excel_timeout = QLineEdit('300.0')
        self.field_major_version = QLineEdit('0')
        self.field_minor_version = QLineEdit('1')
        self.field_add_packages = QLineEdit('["lcm_general", "lcm_math"]')
        self.field_max_parallel_processes = QLineEdit('{:d}'.format(psutil.cpu_count(logical=False)//2))

        self.button_refresh = QPushButton('Refresh Tree', self)
        self.button_depend = QPushButton('Add Dependencies to Selection', self)
        self.button_collect = QPushButton('Collect Details', self)
        self.button_change_version = QPushButton('Change Version of Selection', self)
        self.button_add_packages = QPushButton('Add Packges to Selection', self)

        self.button_build = QPushButton('Build Selected', self)
        self.button_ctest = QPushButton('Build Selected Tests', self)
        self.button_rtest = QPushButton('Run Selected Tests', self)

        self.button_show_build = QPushButton('Show Report', self)
        self.button_show_build_test = QPushButton('Show Report', self)
        self.button_show_run_test = QPushButton('Show Report', self)

        self.button_help_build = QPushButton('Build Help', self)
        self.button_help_show = QPushButton('Show Help', self)

        self.button_excel_refresh = QPushButton('Refresh Tree', self)
        self.button_build_excel_tests = QPushButton('Build Excel Tests', self)
        self.button_run_excel_tests = QPushButton('Run Excel Tests', self)
        self.button_show_build_excel_tests = QPushButton('Show Report', self)
        self.button_show_run_excel_tests = QPushButton('Show Report', self)

        self.check_disabled_build = QCheckBox('Show disabled Builds')
        self.check_collect_sel = QCheckBox('Only for selected')
        self.check_emptyUG = QCheckBox('Delete existing Components')
        self.check_ctest_FreeCAD = QCheckBox('FreeCAD Tests')
        self.check_ctest_Creo = QCheckBox('Creo Tests')
        self.check_run_test_component_run_sim = QCheckBox('Start Simulation')
        self.check_run_test_component_remote = QCheckBox('Remote Compute')

        self.check_help1 = QCheckBox('NitPickyMode')
        self.check_help2 = QCheckBox('OutputOnlyErrors')
        self.check_help3 = QCheckBox('OutputOnlyWarnings')
        self.check_help_web = QCheckBox('Open External')
        self.check_run_excel_test_run_sim = QCheckBox('Start Simulation')
        self.check_run_excel_test_remote = QCheckBox('Remote Compute')
        self.check_emulate_build_server = QCheckBox('Emulate Build Server')

        self.combo_add_packages = QComboBox()
        self.combo_change_version = QComboBox()

        self.initUI()

    def initUI(self):
        self.setWindowTitle(self.title)
        self.setWindowIcon(QtGui.QIcon(self.icon_path))
        # self.setGeometry(self.left, self.top, self.width, self.height)
        self.resize(self.width, self.height)

        '''add tabs'''
        self.tabs.addTab(self.tab1, "Components")
        self.tabs.addTab(self.tab2, "Excel Tests")
        self.tabs.addTab(self.tab3, "Help")

        '''tab 1'''
        self.__addTree()
        self.__refreshTreeAction()
        self.createTreeControlGroupBox()
        self.createControlGroupBox()
        self.createTab1BrowserGroupBox()

        '''tab 2'''
        self.__addExcelTree()
        self.__refreshExcelTreeAction()
        self.createExcelTreeControlGroupBox()
        self.createExcelTestGroupBox()
        self.createTab2BrowserGroupBox()

        '''tab 3'''
        self.createHelpGroupBox()
        self.createTab3BrowserGroupBox()

        main_layout = QVBoxLayout(self)
        self.tab1.layout = QGridLayout()
        self.tab2.layout = QGridLayout()
        self.tab3.layout = QGridLayout()

        '''make sure left (control) and right (browser) column have the same width'''
        self.tab1.layout.setColumnStretch(0, 1)
        self.tab1.layout.setColumnStretch(1, 1)
        self.tab2.layout.setColumnStretch(0, 1)
        self.tab2.layout.setColumnStretch(1, 1)
        self.tab3.layout.setColumnStretch(0, 1)
        self.tab3.layout.setColumnStretch(1, 1)

        '''tab 1'''
        # noinspection PyArgumentList
        self.tab1.layout.addWidget(self.tree, 0, 0)
        # # noinspection PyArgumentList
        # self.tab1.layout.addWidget(self.browser1_nav, 0, 1, 2, 1)
        # noinspection PyArgumentList
        self.tab1.layout.addWidget(self.browser1_box, 0, 1, 3, 1)
        # noinspection PyArgumentList
        self.tab1.layout.addWidget(self.tree_control_box, 1, 0)
        # noinspection PyArgumentList
        self.tab1.layout.addWidget(self.control_box, 2, 0)

        '''tab 2'''
        # noinspection PyArgumentList
        self.tab2.layout.addWidget(self.excel_tree, 0, 0)
        # noinspection PyArgumentList
        self.tab2.layout.addWidget(self.browser2_box, 0, 1, 2, 1)
        # noinspection PyArgumentList
        self.tab2.layout.addWidget(self.excel_tree_control_box, 1, 0)
        # noinspection PyArgumentList
        self.tab2.layout.addWidget(self.excel_box, 2, 0)

        '''tab 3'''
        # noinspection PyArgumentList
        self.tab3.layout.addWidget(self.help_box, 0, 0)
        # noinspection PyArgumentList
        self.tab3.layout.addWidget(self.browser3_box, 0, 1, 2, 1)

        self.tab1.setLayout(self.tab1.layout)
        self.tab2.setLayout(self.tab2.layout)
        self.tab3.setLayout(self.tab3.layout)

        # noinspection PyArgumentList
        main_layout.addWidget(self.tabs)
        self.setLayout(main_layout)

    def createTreeControlGroupBox(self):
        # noinspection PyUnresolvedReferences
        self.button_refresh.clicked.connect(self.__refreshTreeAction)
        self.check_disabled_build.setChecked(False)

        # noinspection PyUnresolvedReferences
        self.button_collect.clicked.connect(self.__collectInfoAction)
        self.check_collect_sel.setChecked(True)

        # noinspection PyUnresolvedReferences
        self.button_change_version.clicked.connect(self.__changeVersionAction)
        # noinspection PyUnresolvedReferences
        self.button_add_packages.clicked.connect(self.__addPackagesAction)

        self.combo_change_version.addItem('Set Version to')
        self.combo_change_version.addItem('Increase Version by')

        self.combo_add_packages.addItem('Automatic Mode')
        self.combo_add_packages.addItem('Custom Mode')

        layout = QGridLayout()

        # noinspection PyArgumentList
        layout.addWidget(self.button_refresh, 0, 0)
        # noinspection PyArgumentList
        layout.addWidget(self.check_disabled_build, 0, 1)
        # noinspection PyArgumentList
        layout.addWidget(self.button_depend, 1, 0)
        # noinspection PyArgumentList
        layout.addWidget(self.button_collect, 2, 0)
        # noinspection PyArgumentList
        layout.addWidget(self.check_collect_sel, 2, 1)
        # noinspection PyArgumentList
        layout.addWidget(self.button_change_version, 3, 0)
        # noinspection PyArgumentList
        layout.addWidget(self.combo_change_version, 3, 1)
        # noinspection PyArgumentList
        layout.addWidget(QLabel('Major'), 3, 2)
        # noinspection PyArgumentList
        layout.addWidget(self.field_major_version, 3, 3)
        # noinspection PyArgumentList
        layout.addWidget(QLabel('Minor'), 3, 4)
        # noinspection PyArgumentList
        layout.addWidget(self.field_minor_version, 3, 5)
        # noinspection PyArgumentList
        layout.addWidget(self.button_add_packages, 4, 0)
        # noinspection PyArgumentList
        layout.addWidget(self.combo_add_packages, 4, 1)
        # noinspection PyArgumentList
        layout.addWidget(QLabel('Custom Packages'), 4, 2)
        # noinspection PyArgumentList
        layout.addWidget(self.field_add_packages, 4, 3)

        self.tree_control_box.setLayout(layout)

    def createExcelTreeControlGroupBox(self):
        # noinspection PyUnresolvedReferences
        self.button_excel_refresh.clicked.connect(self.__refreshExcelTreeAction)

        layout = QGridLayout()

        # noinspection PyArgumentList
        layout.addWidget(self.button_excel_refresh, 0, 0)

        self.excel_tree_control_box.setLayout(layout)

    def createControlGroupBox(self):
        # noinspection PyUnresolvedReferences
        self.button_build.clicked.connect(self.__buildComponentsAction)
        # noinspection PyUnresolvedReferences
        self.button_depend.clicked.connect(self.__selectDependenciesAction)
        # noinspection PyUnresolvedReferences
        self.button_ctest.clicked.connect(self.__buildTestsAction)
        # noinspection PyUnresolvedReferences
        self.button_rtest.clicked.connect(self.__runTestsAction)
        # noinspection PyUnresolvedReferences
        self.button_show_build.clicked.connect(self.__showBuildComponentAction)
        # noinspection PyUnresolvedReferences
        self.button_show_build_test.clicked.connect(self.__showBuildComponentTestAction)
        # noinspection PyUnresolvedReferences
        self.button_show_run_test.clicked.connect(self.__showRunComponentTestAction)

        self.check_emptyUG.setChecked(False)
        self.check_ctest_FreeCAD.setChecked(False)
        self.check_ctest_Creo.setChecked(False)
        self.check_emulate_build_server.setChecked(False)

        layout = QGridLayout()

        # noinspection PyArgumentList
        layout.addWidget(self.button_build, 0, 0)
        # noinspection PyArgumentList
        layout.addWidget(self.check_emptyUG, 0, 1)
        # noinspection PyArgumentList
        layout.addWidget(self.button_show_build, 0, 5)

        # noinspection PyArgumentList
        layout.addWidget(self.button_ctest, 1, 0)
        # noinspection PyArgumentList
        layout.addWidget(self.check_ctest_FreeCAD, 1, 1)
        # noinspection PyArgumentList
        layout.addWidget(self.check_ctest_Creo, 1, 2)
        # noinspection PyArgumentList
        layout.addWidget(self.button_show_build_test, 1, 5)

        # noinspection PyArgumentList
        layout.addWidget(self.button_rtest, 2, 0)
        # noinspection PyArgumentList
        layout.addWidget(self.check_run_test_component_run_sim, 2, 1)
        # noinspection PyArgumentList
        layout.addWidget(self.check_run_test_component_remote, 2, 2)
        # noinspection PyArgumentList
        layout.addWidget(QLabel('Min. Timeout (sec)'), 2, 3)
        # noinspection PyArgumentList
        layout.addWidget(self.field_component_timeout, 2, 4)
        # noinspection PyArgumentList
        layout.addWidget(self.button_show_run_test, 2, 5)

        # noinspection PyArgumentList
        layout.addWidget(QLabel('Common Settings:'), 3, 0)
        # noinspection PyArgumentList
        layout.addWidget(self.check_emulate_build_server, 3, 1)
        # noinspection PyArgumentList
        layout.addWidget(QLabel('Parallel Processes:'), 3, 2)
        # noinspection PyArgumentList
        layout.addWidget(self.field_max_parallel_processes, 3, 3)

        self.control_box.setLayout(layout)

    def createExcelTestGroupBox(self):
        # noinspection PyUnresolvedReferences
        self.button_build_excel_tests.clicked.connect(self.__buildExcelTestsAction)
        # noinspection PyUnresolvedReferences
        self.button_run_excel_tests.clicked.connect(self.__runExcelTestsAction)
        # noinspection PyUnresolvedReferences
        self.button_show_build_excel_tests.clicked.connect(self.__showBuildExcelTestAction)
        # noinspection PyUnresolvedReferences
        self.button_show_run_excel_tests.clicked.connect(self.__showRunExcelTestAction)

        layout = QGridLayout()
        # noinspection PyArgumentList
        layout.addWidget(self.button_build_excel_tests, 0, 0)
        # noinspection PyArgumentList
        layout.addWidget(self.button_show_build_excel_tests, 0, 5)
        # noinspection PyArgumentList
        layout.addWidget(self.button_run_excel_tests, 1, 0)
        # noinspection PyArgumentList
        layout.addWidget(self.check_run_excel_test_run_sim, 1, 1)
        # noinspection PyArgumentList
        layout.addWidget(self.check_run_excel_test_remote, 1, 2)
        # noinspection PyArgumentList
        layout.addWidget(QLabel('Min. Timeout (sec)'), 1, 3)
        # noinspection PyArgumentList
        layout.addWidget(self.field_excel_timeout, 1, 4)
        # noinspection PyArgumentList
        layout.addWidget(self.button_show_run_excel_tests, 1, 5)

        self.excel_box.setLayout(layout)

    def createTab1BrowserGroupBox(self):
        layout = QGridLayout()

        # navigation_bar = QToolBar('Navigation Toolbar')
        # back_button = QAction("Back", self)
        # back_button.setStatusTip('Go to previous page you visited')
        # back_button.triggered.connect(self.tab1_browser.back)
        # navigation_bar.addAction(back_button)
        #
        # refresh_button = QAction("Refresh", self)
        # refresh_button.setStatusTip('Refresh this page')
        # refresh_button.triggered.connect(self.tab1_browser.reload)
        # navigation_bar.addAction(refresh_button)
        #
        # next_button = QAction("Next", self)
        # next_button.setStatusTip('Go to next page')
        # next_button.triggered.connect(self.tab1_browser.forward)
        # navigation_bar.addAction(next_button)

        # layout.addWidget(navigation_bar, 0, 0)

        # noinspection PyArgumentList
        layout.addWidget(self.tab1_browser, 0, 0)
        self.browser1_box.setLayout(layout)

    def createTab2BrowserGroupBox(self):
        layout = QGridLayout()
        # noinspection PyArgumentList
        layout.addWidget(self.tab2_browser, 0, 0)
        self.browser2_box.setLayout(layout)

    def createTab3BrowserGroupBox(self):
        layout = QGridLayout()
        # noinspection PyArgumentList
        layout.addWidget(self.tab3_browser, 0, 0)
        self.browser3_box.setLayout(layout)

    def createHelpGroupBox(self):
        # noinspection PyUnresolvedReferences
        self.button_help_build.clicked.connect(self.__buildHelpAction)
        # noinspection PyUnresolvedReferences
        self.button_help_show.clicked.connect(self.__showHelpAction)

        layout = QGridLayout()
        # noinspection PyArgumentList
        layout.addWidget(self.button_help_build, 0, 0)
        # noinspection PyArgumentList
        layout.addWidget(self.check_help1, 0, 1)
        # noinspection PyArgumentList
        layout.addWidget(self.check_help2, 0, 2)
        # noinspection PyArgumentList
        layout.addWidget(self.check_help3, 0, 3)
        # noinspection PyArgumentList
        layout.addWidget(self.button_help_show, 1, 0)
        # noinspection PyArgumentList
        layout.addWidget(self.check_help_web, 1, 1)

        self.help_box.setLayout(layout)

    def __treeAddDirectory(self, parent, label, key):
        directory = QTreeWidgetItem(parent)
        directory.setText(self.TreeCol, label)
        directory.setText(self.UserGroupCol, '')
        directory.setText(self.DependenciesCol, '')
        directory.setText(self.PackagesCol, '')
        directory.setText(self.VersionCol, '')
        directory.setFlags(directory.flags() | Qt.ItemIsTristate | Qt.ItemIsUserCheckable)
        directory.setCheckState(self.TreeCol, Qt.Unchecked)
        self.tree_directories[key] = directory

    def __treeAddComponent(self, parent, label, key, user_group='', packages='', version=''):
        comp = QTreeWidgetItem(parent)
        comp.setText(self.TreeCol, label)
        comp.setText(self.UserGroupCol, user_group)
        comp.setText(self.DependenciesCol, '')
        comp.setText(self.PackagesCol, packages)
        comp.setText(self.VersionCol, version)
        comp.setFlags(comp.flags() | Qt.ItemIsTristate | Qt.ItemIsUserCheckable)
        comp.setCheckState(self.TreeCol, Qt.Unchecked)
        self.tree_components[key] = comp

    def __addTree(self):
        header = self.tree.headerItem()
        header.setText(self.TreeCol, 'Tree')
        header.setText(self.UserGroupCol, 'UserGroup')
        header.setText(self.DependenciesCol, 'Dependencies')
        header.setText(self.PackagesCol, 'Packages')
        header.setText(self.VersionCol, 'Version')
        self.tree.setColumnWidth(self.TreeCol, 400)

        self.tree.setContextMenuPolicy(Qt.CustomContextMenu)
        # noinspection PyUnresolvedReferences
        self.tree.customContextMenuRequested.connect(self.__menuContextTree)

    def __menuContextTree(self, point):
        index = self.tree.indexAt(point)

        if not index.isValid():
            return

        menu = QMenu()
        open_action = menu.addAction("Open Directory")

        action = menu.exec_(self.tree.mapToGlobal(point))

        if action is not None:
            if action == open_action:
                item = self.tree.itemAt(point)
                key = self.findDirectoryInTree(item)
                if key is None:
                    key = self.findComponentInTree(item)
                if key is not None:
                    path = self.keyToFullPath(key)
                    self.__openDirectory(path)
                else:
                    print('Unable to find directory')

    def __excelTreeAddDirectory(self, parent, label, key):
        directory = QTreeWidgetItem(parent)
        directory.setText(self.TreeCol, label)
        directory.setFlags(directory.flags() | Qt.ItemIsTristate | Qt.ItemIsUserCheckable)
        directory.setCheckState(self.TreeCol, Qt.Unchecked)
        self.excel_tree_directories[key] = directory

    def __excelTreeAddFile(self, parent, label, key):
        file = QTreeWidgetItem(parent)
        file.setText(self.TreeCol, label)
        file.setFlags(file.flags() | Qt.ItemIsTristate | Qt.ItemIsUserCheckable)
        file.setCheckState(self.TreeCol, Qt.Unchecked)
        self.excel_tree_files[key] = file

    def __addExcelTree(self):
        header = self.excel_tree.headerItem()
        header.setText(self.TreeCol, 'Tree')
        self.excel_tree.setColumnWidth(self.TreeCol, 400)

    def __openDirectory(self, dirpath):
        """Open the function directory in the system file manager"""
        try:
            real_path = os.path.realpath(dirpath)
            if sys.platform == "win32":
                '''Windows'''
                subprocess.call("explorer " + real_path)
                # subprocess.call("cmd /c start " + real_path)
            elif sys.platform == "linux" or sys.platform == "linux2":
                '''linux'''
                raise NotImplementedError
            elif sys.platform == "darwin":
                ''' OS X'''
                raise NotImplementedError
        except Exception as e:
            print('Unable to open directory')

    @try_catch
    def __refreshTreeAction(self, *args, **kwargs):
        if os.path.isfile(self.info_file):
            info = json.loads(open(self.info_file).read())
        else:
            info = {}

        self.tree.clear()
        self.tree_directories = {}
        self.tree_components = {}

        if self.check_disabled_build.isChecked():
            file_name = 'disabled_build_component.py'
        else:
            file_name = 'build_component.py'

        file_paths = fu.searchBuildComponentScripts(cs_root, file_name=file_name)

        parent_paths = []
        for abs_path in file_paths:
            rel_path = os.path.relpath(abs_path, cs_root)
            path_list = rel_path.split(os.path.sep)

            for index in range(1, len(path_list)):
                dir_path = os.path.sep.join(path_list[:-index])
                parent_paths.append(dir_path)

        '''make sure every directory appears only once in the list'''
        parent_keys = list(set(parent_paths))
        parent_keys.sort()

        for key in parent_keys:
            key_list = key.split(os.path.sep)
            full_path = os.path.join(cs_root, key)
            build_script = os.path.join(full_path, file_name)

            if len(key_list) == 1:
                parent = self.tree
                label = key_list[0]
            else:
                parent_key = os.path.sep.join(key_list[:-1])
                parent = self.tree_directories[parent_key]
                label = key_list[-1]

            if os.path.isfile(build_script):
                user_group = ''
                packages = ''
                version = ''
                if full_path in info:
                    user_group = info[full_path]['UserGroup'] if 'UserGroup' in info[full_path] else ''
                    packages = ';'.join(info[full_path]['Packages'].keys()) if 'Packages' in info[full_path] else ''
                    version = info[full_path]['Version'] if 'Version' in info[full_path] else ''

                self.__treeAddComponent(parent, label, key, user_group=user_group, packages=packages, version=version)
            else:
                self.__treeAddDirectory(parent, label, key)

    @try_catch
    def __refreshExcelTreeAction(self, *args, **kwargs):
        self.excel_tree.clear()

        file_paths = glob.glob(cs_root + '/**/*.sym.xlsx', recursive=True)

        parent_paths = []
        rel_file_paths = []
        for abs_path in file_paths:
            rel_path = os.path.relpath(abs_path, cs_root)
            rel_file_paths.append(rel_path)
            path_list = rel_path.split(os.path.sep)

            for index in range(1, len(path_list)):
                dir_path = os.path.sep.join(path_list[:-index])
                parent_paths.append(dir_path)

        '''make sure every directory appears only once in the list'''
        parent_keys = list(set(parent_paths))
        parent_keys.sort()

        for key in parent_keys:
            key_list = key.split(os.path.sep)

            if len(key_list) == 1:
                parent = self.excel_tree
                label = key_list[0]
            else:
                parent_key = os.path.sep.join(key_list[:-1])
                parent = self.excel_tree_directories[parent_key]
                label = key_list[-1]

            self.__excelTreeAddDirectory(parent, label, key)

        for file in rel_file_paths:
            key_list = file.split(os.path.sep)
            parent_key = os.path.sep.join(key_list[:-1])
            parent = self.excel_tree_directories[parent_key]
            label = os.path.basename(file)

            self.__excelTreeAddFile(parent, label, file)

    @try_catch
    def __selectDependenciesAction(self, *args, **kwargs):
        key_list = []
        for key, item in self.tree_components.items():
            if item.checkState(self.TreeCol) > 0:
                key_list.append('%ComponentSpace%' + os.path.sep + key)

        d = json.loads(open(os.path.join(cs_root, 'build_dependencies.json')).read())

        for key in key_list:
            if key in d:
                for element in d[key]:
                    tree_key = element.replace('%ComponentSpace%' + os.path.sep, '')
                    if tree_key in self.tree_components:
                        item = self.tree_components[tree_key]
                        item.setCheckState(self.TreeCol, Qt.Checked)

    @try_catch
    def __collectInfoAction(self, *args, **kwargs):
        if self.check_collect_sel.isChecked():
            components_list = self.getSelectedComponents()
        else:
            components_list = None

        json_file = self.__writeJSONFile(components_list)

        options = {'--function': 'collect_info', '--file': json_file}
        csymspaceRun(self.script, timeout=300, stdout=sys.stdout, stderr=sys.stderr, cmd_kwargs=options)

    @try_catch
    def __changeVersionAction(self, *args, **kwargs):
        components_list = self.getSelectedComponents()

        json_file = self.__writeJSONFile(components_list)
        major = str(self.field_major_version.text())
        minor = str(self.field_minor_version.text())

        text = self.combo_change_version.currentText().lower()
        if text == 'increase version by':
            increase_version = '1'  # True
        elif text == 'set version to':
            increase_version = '0'  # False
        else:
            raise NotImplementedError('Invalid option: {}'.format(text))

        options = {'--function': 'change_version', '--file': json_file, '--major': major, '--minor': minor,
                   '--increase_version': increase_version}
        csymspaceRun(self.script, timeout=300, stdout=sys.stdout, stderr=sys.stderr, cmd_kwargs=options)

    @try_catch
    def __addPackagesAction(self, *args, **kwargs):
        components_list = self.getSelectedComponents()

        try:
            mode = self.combo_add_packages.currentText().lower()
            if mode == 'custom mode':
                auto_collect_pkgs = False
                text = str(self.field_add_packages.text())
                packages = ast.literal_eval(text)

                if isinstance(packages, str):
                    '''scalar to list'''
                    packages = [packages, ]
            elif mode == 'automatic mode':
                auto_collect_pkgs = True
                packages = []
            else:
                raise NotImplementedError('Invalid option: {}'.format(mode))

            json_file = self.__writeJSONFile(components_list)
            pkg_file = self.__writeJSONFile(packages)

            options = {'--function': 'add_packages', '--file': json_file,
                       '--pkg_file': pkg_file, '--auto_collect_pkgs': str(int(auto_collect_pkgs))}
            csymspaceRun(self.script, timeout=300, stdout=sys.stdout, stderr=sys.stderr, cmd_kwargs=options)
        except Exception as ex:
            print(ex)

    @try_catch
    def __buildComponentsAction(self, *args, **kwargs):
        components_list = self.getSelectedComponents()
        empty_usergroups = str(int(self.check_emptyUG.isChecked()))
        emulate_build_server = str(int(self.check_emulate_build_server.isChecked()))

        json_file = self.__writeJSONFile(components_list)
        options = {'--function': 'build_components', '--file': json_file,
                   '--emptyUserGroups': empty_usergroups, '--emulate_build_server': emulate_build_server}

        pythonRun(self.script, timeout=3600.0, stdout=sys.stdout, stderr=sys.stderr, cmd_kwargs=options)

    @try_catch
    def __buildTestsAction(self, *args, **kwargs):
        components_list = self.getSelectedComponents()
        build_freecad = str(int(self.check_ctest_FreeCAD.isChecked()))
        build_creo = str(int(self.check_ctest_Creo.isChecked()))
        emulate_build_server = str(int(self.check_emulate_build_server.isChecked()))

        json_file = self.__writeJSONFile(components_list)
        options = {'--function': 'build_component_tests', '--file': json_file,
                   '--build_freecad': build_freecad, '--build_creo': build_creo,
                   '--emulate_build_server': emulate_build_server}

        pythonRun(self.script, timeout=3600.0, stdout=sys.stdout, stderr=sys.stderr, cmd_kwargs=options)

    @try_catch
    def __buildExcelTestsAction(self, *args, **kwargs):
        excel_list = self.getSelectedExcelTests()
        json_file = self.__writeJSONFile(excel_list)
        options = {'--function': 'build_excel_tests', '--file': json_file}
        pythonRun(self.script, timeout=3600.0, stdout=sys.stdout, stderr=sys.stderr, cmd_kwargs=options)

    @try_catch
    def __runExcelTestsAction(self, *args, **kwargs):
        excel_list = self.getSelectedExcelTests()
        json_file = self.__writeJSONFile(excel_list)

        run_sim = str(int(self.check_run_excel_test_run_sim.isChecked()))
        remote_compute = str(int(self.check_run_excel_test_remote.isChecked()))
        timeout = str(self.field_excel_timeout.text())
        options = {'--function': 'run_excel_tests', '--file': json_file,
                   '--run_sim': run_sim, '--remote_compute': remote_compute, '--timeout': timeout}
        pythonRun(self.script, timeout=3600.0, stdout=sys.stdout, stderr=sys.stderr, cmd_kwargs=options)

    @try_catch
    def __runTestsAction(self, *args, **kwargs):
        components_list = self.getSelectedComponents()
        json_file = self.__writeJSONFile(components_list)

        run_sim = str(int(self.check_run_test_component_run_sim.isChecked()))
        remote_compute = str(int(self.check_run_test_component_remote.isChecked()))
        timeout = str(self.field_component_timeout.text())
        emulate_build_server = str(int(self.check_emulate_build_server.isChecked()))
        n_max_proc = str(self.field_max_parallel_processes.text())

        options = {'--function': 'run_component_tests', '--file': json_file,
                   '--run_sim': run_sim, '--remote_compute': remote_compute, '--timeout': timeout,
                   '--emulate_build_server': emulate_build_server, '--n_max_proc': n_max_proc}
        pythonRun(self.script, timeout=3600.0, stdout=sys.stdout, stderr=sys.stderr, cmd_kwargs=options)

    @try_catch
    def __showBuildComponentAction(self, *args, **kwargs):
        self.__showJUnitReport(ComponentBuilder.build_reports_dir, show_matrix=True, open_external=False, tab=1)

    @try_catch
    def __showBuildComponentTestAction(self, *args, **kwargs):
        self.__showJUnitReport(ComponentTestBuilder.build_reports_dir, show_matrix=True, open_external=False, tab=1)

    @try_catch
    def __showRunComponentTestAction(self, *args, **kwargs):
        self.__showJUnitReport(ComponentTestBuilder.run_reports_dir, show_matrix=True, open_external=False, tab=1)

    @try_catch
    def __showBuildExcelTestAction(self, *args, **kwargs):
        self.__showJUnitReport(ProjectTestBuilder.build_reports_dir, show_matrix=True, open_external=False, tab=2)

    @try_catch
    def __showRunExcelTestAction(self, *args, **kwargs):
        self.__showJUnitReport(ProjectTestBuilder.run_reports_dir, show_matrix=True, open_external=False, tab=2)

    def __showJUnitReport(self, output_dir, show_matrix, open_external, tab=1):
        junit_files = glob.glob(output_dir + '/**/*-junit.xml', recursive=True)

        if len(junit_files) > 0:
            if show_matrix:
                html_file = junit2html.junit2htmlMatrix(junit_files[0], show=open_external)
            else:
                html_file = junit2html.junit2html(junit_files[0], show=open_external)
            if not open_external:
                url = pathlib.Path(html_file).as_uri()
                if tab == 1:
                    self.tab1_browser.setUrl(QUrl(url))
                else:
                    self.tab2_browser.setUrl(QUrl(url))
        else:
            print('No Test reports found')

        print('Finished')

    @try_catch
    def __buildHelpAction(self, *args, **kwargs):
        runner = DeveloperHelp()
        runner.setNitPickyMode(self.check_help1.isChecked())
        runner.setOutputOnlyErrors(self.check_help2.isChecked())
        runner.setOutputOnlyWarnings(self.check_help3.isChecked())
        thread = threading.Thread(target=runner.build, kwargs=dict(sys_exit=False))
        thread.start()

    @try_catch
    def __showHelpAction(self, *args, **kwargs):
        dev_help = DeveloperHelp()
        if self.check_help_web.isChecked():
            dev_help.showOutputHTML()
        else:
            url = pathlib.Path(dev_help.html_file).as_uri()
            self.tab3_browser.setUrl(QUrl(url))

    @staticmethod
    def iter_tree_widget(root):
        iterator = QTreeWidgetItemIterator(root)
        while True:
            item = iterator.value()
            if item is not None:
                yield item
                iterator += 1
            else:
                break

    def checkState(self):
        column = self.TreeCol
        # for item in self.iter_tree_widget(self.tree):
        for key, item in self.tree_components.items():
            print('State: %s, Text: "%s"' % (
                item.checkState(column), item.text(column)))

    def getSelectedComponents(self):
        components_list = []
        for key, item in self.tree_components.items():
            if item.checkState(self.TreeCol) > 0:
                components_list.append(self.keyToFullPath(key))
        return components_list

    def getSelectedExcelTests(self):
        excel_list = []
        for key, item in self.excel_tree_files.items():
            if item.checkState(self.TreeCol) > 0:
                excel_list.append(self.keyToFullPath(key))
        return excel_list

    def findComponentInTree(self, item):
        for key, value in self.tree_components.items():
            if value == item:
                return key
        return None

    def findDirectoryInTree(self, item):
        for key, value in self.tree_directories.items():
            if value == item:
                return key
        return None

    @staticmethod
    def keyToFullPath(key):
        return os.path.join(cs_root, key)


if __name__ == '__main__':
    _app = QApplication([])
    
    try:
        _cm = ComponentSpaceCenterQt()
        _cm.show()
        _return_code = _app.exec_()
    except Exception as _ex:
        showErrorInGUI(_ex)
        _return_code = 1

    sys.exit(_return_code)