#!/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision: 4482 $
$LastChangedDate:: 2021-06-15 11:04:42 +0200#$

This file is part of SyMSpace.

Copyright (c) 2018, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""

import os
import sys
import webbrowser
import optparse
import sphinx.cmd.build
from symspace.osutils.pathutils import CMDParser


class SphinxHelp:

    @classmethod
    def fromCMDParser(cls) -> 'SphinxHelp':
        """
        Commandline parser to create instance of this class

        Commandline Arguments
        ---------------------

        Todo: add docstrings

        Returns
        -------
        sphinx_help : SphinxHelp
        """

        parser = optparse.OptionParser("usage: %prog [options]")
        parser.add_option("--SrcDir", dest="source_dir",
                          default="", type="string",
                          help="RST source directory")
        parser.add_option("--OutDir", dest="output_dir", default="",
                          type="string", help="Sphinx output directory")
        parser.add_option("--WarningsAsErrors", dest="warnings_as_erros", default=0,
                          type="int", help="Set warnings as errors")
        parser.add_option("--OutputOnlyWarnings", dest="output_only_warnings", default=0,
                          type="int", help="Output only warnings")
        parser.add_option("--OutputOnlyErrors", dest="output_only_errors", default=0,
                          type="int", help="Output only errors")
        parser.add_option("--NitPickyMode", dest="nitpicky_mode", default=0,
                          type="int", help="Set nit-picky mode")
        parser.add_option("--LogFile", dest="log_file", default="0",
                          type="string", help="Log file path")

        args = CMDParser.jepSysArgs()
        (options, args) = parser.parse_args(args)
        if len(args) != 0:
            parser.error("incorrect number of arguments")

        if options.source_dir != "":
            source_dir = CMDParser.jepArgPath(options.source_dir)
        else:
            source_dir = None
        if options.output_dir != "":
            output_dir = CMDParser.jepArgPath(options.output_dir)
        else:
            output_dir = None

        if options.log_file != "0":
            log_file = CMDParser.jepArgPath(options.log_file)
        else:
            log_file = None

        sphinx_help = cls(source_dir, output_dir)

        sphinx_help.setWarningsAsErrors(bool(options.warnings_as_erros))
        sphinx_help.setOutputOnlyWarnings(bool(options.output_only_warnings))
        sphinx_help.setOutputOnlyErrors(bool(options.output_only_errors))
        sphinx_help.setNitPickyMode(bool(options.nitpicky_mode))
        sphinx_help.setLogFile(log_file)
        return sphinx_help

    def __init__(self, source_dir: str, output_dir: str):

        self.source_dir = source_dir
        self.output_dir = output_dir
        self.html_file = os.path.join(self.output_dir, 'index.html')
        self.build_options = ['-b', 'html', self.source_dir, self.output_dir]

    def showOutputHTML(self):
        webbrowser.open('file://' + self.html_file, new=2)

    @staticmethod
    def writeHead():
        pass

    def setLogFile(self, log_file_path: str):
        """
        Write warnings (and errors) to the given file, in addition to standard error.

        Parameters
        ----------
        log_file_path : str

        """
        if (log_file_path != '') and (log_file_path is not None):
            # if log_file_path == 'ComponentSpace':
            #     log_file_path = os.path.join(os.getenv('ComponentSpace'), 'help.log')

            log_file_dir = os.path.dirname(log_file_path)
            if not os.path.isdir(log_file_dir):
                '''create directory if path does not already exist'''
                os.makedirs(log_file_dir)

            print('Log File: {}'.format(log_file_path))
            self.build_options.append('-w')
            self.build_options.append(log_file_path)
            self.setSwitch('-q', True)

    def setNitPickyMode(self, on):
        """ Run in nit-picky mode. Currently, this generates warnings for all missing references.

        Parameters
        ----------
        on : bool
            if true the switch will be set else it might be removed
        """
        print('Set nit-picky mode: {}'.format('ON' if on else 'OFF'))
        self.setSwitch('-n', on)

    def setWarningsAsErrors(self, on):
        """
        Turn warnings into errors.

        Parameters
        ----------
        on : bool
            if true the switch will be set else it might be removed
        """
        print('Turn warnings into errors: {}'.format('ON' if on else 'OFF'))
        self.setSwitch('-W', on)

    def setOutputOnlyWarnings(self, on):
        """
        Do not output anything on standard output, only write warnings and errors to standard error.

        Parameters
        ----------
        on : bool
            if true the switch will be set else it might be removed
        """
        print('Write only warnings and errors to output: {}'.format('ON' if on else 'OFF'))
        self.setSwitch('-q', on)

    def setOutputOnlyErrors(self, on):
        """
        Do not output anything on standard output, also suppress warnings. Only errors are written to standard error.

        Parameters
        ----------
        on : bool
            if true the switch will be set else it might be removed
        """
        print('Write only errors to output: {}'.format('ON' if on else 'OFF'))
        self.setSwitch('-Q', on)

    def increaseVerbosity(self, on):
        """
        Increase verbosity (loglevel). This option can be given up to three times to get more debug logging output.

        Parameters
        ----------
        on : bool
            if true the switch will be set else it might be removed
        """
        if on:
            print('Increased verbosity')
        else:
            print('Decreased verbosity')
        self.setSwitch('-v', on)

    def build(self, sys_exit: bool = False):
        self.writeHead()

        val = sphinx.cmd.build.build_main(self.build_options)

        if val == 0:
            print('\nSphinx documentation finished successfully\n\n')
        else:
            print('\nSphinx documentation finished with error, retVal: {}\n\n'.format(val))
        if sys_exit:
            sys.exit(val)

    def setSwitch(self, key, on):
        if on:
            self.build_options.append(key)
        else:
            if key in self.build_options:
                self.build_options.remove(key)
